#load "graphics.cma";;
#load "str.cma";;
Graphics.open_graph " 500x500";;

module type DATA =
sig
  exception Empty
  type 'a e = (float array * 'a)
  type 'a t
  val create : unit -> 'a t
  val add :  'a e -> 'a t -> unit
  val fold : ('a -> 'b e -> 'a ) -> 'a -> 'b t -> 'a
  val size : 'a t -> int
  val sample : 'a t -> 'a e
end

module UtilsData = functor (Data:DATA) -> struct
  include Data

  let iter f data = 
    let () = Data.fold (fun () e -> f e) () data in ()

  let image data =
    let h = Hashtbl.create 10 in
    iter (fun (_,e) ->
      if not (Hashtbl.mem h e) then Hashtbl.add h e 1
      else Hashtbl.replace h e (1+(Hashtbl.find h e)))
      data;
    fst (Hashtbl.fold (fun x _ (l,i) -> ((x,i)::l,i+1)) h ([],0))

  let dim data =
    try Array.length (fst (Data.sample data)) with Empty -> 0

  open Graphics
  let plot_point x y (minx,maxx,miny,maxy) = 
    let xi = 10 + int_of_float (((x-.minx) /. (maxx-.minx))*.(float (size_x ()) -. 20.0))
    and yi = 10 + int_of_float (((y-.miny) /. (maxy-.miny))*.(float (size_y ()) -. 20.0)) in
    draw_circle xi yi 2

  let minmax ?(d1=0) ?(d2=1) d =
    Data.fold (fun (min1x,max1x,min1y,max1y) (da,_) ->
      (min min1x da.(d1),max max1x da.(d1),
       min min1y da.(d2),max max1y da.(d2)))
      (let (da,_) = Data.sample d in (da.(d1), da.(d1), da.(d2), da.(d2)))
      d

  let draw ?(d1=0) ?(d2=1) d color = 
    clear_graph ();
    Data.fold (fun () (da,c) -> 
      set_color (color c);
      plot_point da.(d1) da.(d2) (minmax ~d1 ~d2 d)) () d

  let data_of_list l c =
    let t = Array.create (List.length l -1) 0.0 in
    let rec fill i = function 
      | [] -> failwith "emptyline"
      | [v] -> c v
      | tv::q -> t.(i)<-float_of_string tv; fill (i+1) q in
    let v = fill 0 l in (t,v)

  let load fp c =
    let blank = Str.regexp " \\|\t" in
    let f = open_in fp in
    let d = Data.create () in
    (try while true do
	let line = input_line f in
	Data.add (data_of_list (Str.split blank line) c) d;
      done with End_of_file -> ());
    close_in f;
    d 
end  

module Data1:DATA = struct
  exception Empty
  type 'a e = (float array * 'a)
  type 'a t = (('a e) list) ref
   
  let create () = ref []
  let add a b = b := a::(!b)
  let fold f e t = List.fold_left f e !t
  let size t = List.length !t
  let sample d = try List.hd !d with Failure _ -> raise Empty
end

module Data2:DATA = struct
  exception Empty
  type 'a e = (float array * 'a)
  type 'a t = { mutable table:('a e) array option;
		mutable size:int
	      }
   
  let create () = {table = None; size=0}
  let rec add a t = match t.table with
      None -> t.table <- Some(Array.create 10 a);
	t.size <- 1
    | Some(t1) when Array.length t1 > t.size ->
      t1.(t.size) <- a;
      t.size <- t.size +1
    | Some(t1) -> let t2 = Array.create (2*t.size) t1.(0) in
		  Array.blit t1 0 t2 0 t.size;
		  t.table <- Some(t2);
		  add a t
	
  let fold f e t = match t.table with
      None -> e
    | Some t1 -> let buff = ref e in
		 for i = t.size downto 0 do
		   buff := f !buff t1.(i);
		 done;
		 !buff
  let size t = t.size
  let sample d = match d.table with
      Some(t) -> t.(0)
    | None -> raise Empty
end


module Data = UtilsData(Data2)

module DTree = functor (Data:DATA) -> struct
  open Graphics
  type 'a tree =
      Leaf of 'a
    | Node of (('a tree)*int*float*('a tree));;
  
  let draw1 fl t = 
    let rec draw_rec x y w h= function
      | Leaf(lv) -> lineto x y; fl lv
      | Node(fg,d,v,fd) -> 
	lineto x y;
	draw_rec (x-w) (y-h) (w/2) h fg;
	moveto x y;
	draw_rec (x+w) (y-h) (w/2) h fd;
	let s = Printf.sprintf "x%i <= %f" (d+1) v in
	let st,sth = text_size s in
	set_color white;
	fill_rect (x-st/2) y st sth;
	set_color black;
	moveto (x-st/2) y;
	draw_string s in
    moveto ((size_x ())/2) ((size_x ())-20);
    draw_rec ((size_x ())/2) ((size_x ())-20) ((size_x ())/4) 30 t 

  let draw2 fl ?(d1=0) ?(d2=1) minmax d =
    let rec draw2_rec ?(d1=0) ?(d2=1) (minx,maxx,miny,maxy) (minx2,maxx2,miny2,maxy2) = function 
      | Leaf(x) -> let () = fl (minx2,maxx2,miny2,maxy2) x in ()
      | Node(fg,d,v,fd) -> 

	if d = d1 then
	  let x = 10 + int_of_float (((v-.minx) /. (maxx-.minx))*.(float (size_x ()) -. 20.0)) in
	  moveto x miny2;
	  lineto x maxy2;
	  draw2_rec ~d1 ~d2 (minx,maxx,miny,maxy) (minx2,x,miny2,maxy2) fg;
	  draw2_rec ~d1 ~d2 (minx,maxx,miny,maxy) (x,maxx2,miny2,maxy2) fd;
	else if d=d2 then
	  let y = 10 + int_of_float (((v-.miny) /. (maxy-.miny))*.(float (size_y ()) -. 20.0)) in
	  moveto minx2 y;
	  lineto maxx2 y;
	  draw2_rec ~d1 ~d2 (minx,maxx,miny,maxy) (minx2,maxx2,miny2,y) fg;
	  draw2_rec ~d1 ~d2 (minx,maxx,miny,maxy) (minx2,maxx2,y,maxy2) fd;
	else (
	  draw2_rec ~d1 ~d2 (minx,maxx,miny,maxy) (minx2,maxx2,miny2,maxy2) fg;
	  draw2_rec ~d1 ~d2 (minx,maxx,miny,maxy) (minx2,maxx2,miny2,maxy2) fd
	) in
    draw2_rec ~d1 ~d2 minmax (0,size_x (),0,size_y ()) d

  let fold_data data (f:'c -> 'a Data.e-> 'c) tree init fin =
    let rec checkvar v = function
  | [] -> true
  | (d,vt,true)::q ->
    if v.(d) <= vt then checkvar v q else false
  | (d,vt,false)::q -> 
    if v.(d) > vt then checkvar v q else false in
    let rec fold_rec l = function
  | Node(fg,d,v,fd) ->
    Node(fold_rec ((d,v,true)::l) fg,d,v,
	 fold_rec ((d,v,false)::l) fd)
  | Leaf(lv) -> 
    let re = Data.fold (
      fun c (v,e) ->
	if checkvar v l then f c (v,e)
	else c) 
      (init lv) data in
    Leaf(fin re) in
    fold_rec [] tree   
end

module DecisionTree = DTree(Data)
